@@ -146,6 +146,12 @@ struct mapent_cache {
 	struct mapent **hash;
 };
 
+struct stack {
+	char *mapent;
+	time_t age;
+	struct stack *next;
+};
+
 struct mapent {
 	struct mapent *next;
 	struct list_head ino_index;
@@ -159,6 +165,7 @@ struct mapent {
 	struct mapent *parent;
 	char *key;
 	char *mapent;
+	struct stack *stack;
 	time_t age;
 	/* Time of last mount fail */
 	time_t status;
@@ -175,6 +182,8 @@ void cache_readlock(struct mapent_cache *mc);
 void cache_writelock(struct mapent_cache *mc);
 int cache_try_writelock(struct mapent_cache *mc);
 void cache_unlock(struct mapent_cache *mc);
+int cache_push_mapent(struct mapent *me, char *mapent);
+int cache_pop_mapent(struct mapent *me);
 struct mapent_cache *cache_init(struct autofs_point *ap, struct map_source *map);
 struct mapent_cache *cache_init_null_cache(struct master *master);
 int cache_set_ino_index(struct mapent_cache *mc, const char *key, dev_t dev, ino_t ino);
@@ -189,6 +198,7 @@ struct mapent *cache_lookup_offset(const char *prefix, const char *offset, int s
 struct mapent *cache_partial_match(struct mapent_cache *mc, const char *prefix);
 int cache_add(struct mapent_cache *mc, struct map_source *ms, const char *key, const char *mapent, time_t age);
 int cache_update_offset(struct mapent_cache *mc, const char *mkey, const char *key, const char *mapent, time_t age);
+void cache_update_negative(struct mapent_cache *mc, struct map_source *ms, const char *key, time_t timeout);
 int cache_set_parents(struct mapent *mm);
 int cache_update(struct mapent_cache *mc, struct map_source *ms, const char *key, const char *mapent, time_t age);
 int cache_delete(struct mapent_cache *mc, const char *key);
@@ -233,7 +243,7 @@ int rmdir_path(struct autofs_point *ap, const char *path, dev_t dev);
 #define AUTOFS_LOOKUP_VERSION 5
 
 #define KEY_MAX_LEN    NAME_MAX
-#define MAPENT_MAX_LEN 4095
+#define MAPENT_MAX_LEN 16384
 #define PARSE_MAX_BUF	KEY_MAX_LEN + MAPENT_MAX_LEN + 2
 
 int lookup_nss_read_master(struct master *master, time_t age);
@@ -334,6 +344,7 @@ size_t _strlen(const char *str, size_t max);
 int cat_path(char *buf, size_t len, const char *dir, const char *base);
 int ncat_path(char *buf, size_t len,
               const char *dir, const char *base, size_t blen);
+int _strncmp(const char *s1, const char *s2, size_t n);
 
 /* Core automount definitions */
 
@@ -455,6 +466,9 @@ struct kernel_mod_version {
 /* Don't use bind mounts even when system supports them */
 #define MOUNT_FLAG_NOBIND		0x0020
 
+/* Use symlinks instead of bind mounting local mounts */
+#define MOUNT_FLAG_SYMLINK		0x0040
+
 struct autofs_point {
 	pthread_t thid;
 	char *path;			/* Mount point name */
@@ -512,8 +526,8 @@ int handle_packet_expire_indirect(struct autofs_point *ap, autofs_packet_expire_
 int handle_packet_expire_direct(struct autofs_point *ap, autofs_packet_expire_direct_t *pkt);
 int handle_packet_missing_indirect(struct autofs_point *ap, autofs_packet_missing_indirect_t *pkt);
 int handle_packet_missing_direct(struct autofs_point *ap, autofs_packet_missing_direct_t *pkt);
-void rm_unwanted(unsigned logopt, const char *path, int incl, dev_t dev);
-int count_mounts(unsigned logopt, const char *path, dev_t dev);
+void rm_unwanted(struct autofs_point *ap, const char *path, int incl);
+int count_mounts(struct autofs_point *ap, const char *path, dev_t dev);
 
 #define mounts_mutex_lock(ap) \
 do { \
@@ -544,7 +558,8 @@ static inline void check_cloexec(int fd)
 {
 	if (cloexec_works == 0) {
 		int fl = fcntl(fd, F_GETFD);
-		cloexec_works = (fl & FD_CLOEXEC) ? 1 : -1;
+		if (fl != -1)
+			cloexec_works = (fl & FD_CLOEXEC) ? 1 : -1;
 	}
 	if (cloexec_works > 0)
 		return;
