@@ -302,8 +302,10 @@ static int read_one(unsigned logopt, FILE *f, char *key, unsigned int *k_len, ch
 				if (gotten == got_real || gotten == getting)
 					goto got_it;
 			} else if (mapent_len < MAPENT_MAX_LEN) {
-				mapent_len++;
-				*(p++) = ch;
+				if (p) {
+					mapent_len++;
+					*(p++) = ch;
+				}
 				nch = getc(f);
 				if (nch == EOF &&
 				   (gotten == got_real || gotten == getting))
@@ -397,8 +399,9 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 	unsigned int path_len, ent_len;
 	int entry, cur_state;
 
+	/* Don't return fail on self include, skip source */
 	if (master->recurse)
-		return NSS_STATUS_UNAVAIL;
+		return NSS_STATUS_TRYAGAIN;
 
 	if (master->depth > MAX_INCLUDE_DEPTH) {
 		error(logopt, MODPREFIX
@@ -443,7 +446,7 @@ int lookup_read_master(struct master *master, time_t age, void *context)
 
 			inc = check_master_self_include(master, ctxt);
 			if (inc) 
-				master->recurse = 1;;
+				master->recurse = 1;
 			master->depth++;
 			status = lookup_nss_read_master(master, age);
 			if (!status) {
@@ -526,21 +529,17 @@ static int check_self_include(const char *key, struct lookup_context *ctxt)
 
 static struct map_source *
 prepare_plus_include(struct autofs_point *ap,
+		     struct map_source *source,
 		     time_t age, char *key, unsigned int inc,
 		     struct lookup_context *ctxt)
 {
-	struct map_source *current;
-	struct map_source *source;
+	struct map_source *new;
 	struct map_type_info *info;
 	const char *argv[2];
 	char **tmp_argv, **tmp_opts;
 	int argc;
 	char *buf;
 
-	current = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
 	/*
 	 * TODO:
 	 * Initially just consider the passed in key to be a simple map
@@ -593,21 +592,21 @@ prepare_plus_include(struct autofs_point *ap,
 	}
 	argc += ctxt->opts_argc;
 
-	source = master_find_source_instance(current,
-					     info->type, info->format,
-					     argc, (const char **) tmp_argv);
-	if (source) {
+	new = master_find_source_instance(source,
+					  info->type, info->format,
+					  argc, (const char **) tmp_argv);
+	if (new) {
 		/*
 		 * Make sure included map age is in sync with its owner
 		 * or we could incorrectly wipe out its entries.
 		 */
-		source->age = age;
-		source->stale = 1;
+		new->age = age;
+		new->stale = 1;
 	} else {
-		source = master_add_source_instance(current,
-						    info->type, info->format, age,
-						    argc, (const char **) tmp_argv);
-		if (!source) {
+		new = master_add_source_instance(source,
+						 info->type, info->format, age,
+						 argc, (const char **) tmp_argv);
+		if (!new) {
 			free_argv(argc, (const char **) tmp_argv);
 			free_map_type_info(info);
 			free(buf);
@@ -617,14 +616,14 @@ prepare_plus_include(struct autofs_point *ap,
 	}
 	free_argv(argc, (const char **) tmp_argv);
 
-	source->depth = current->depth + 1;
+	new->depth = source->depth + 1;
 	if (inc)
-		source->recurse = 1;
+		new->recurse = 1;
 
 	free_map_type_info(info);
 	free(buf);
 
-	return source;
+	return new;
 }
 
 int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
@@ -645,7 +644,7 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 	mc = source->mc;
 
 	if (source->recurse)
-		return NSS_STATUS_UNAVAIL;
+		return NSS_STATUS_TRYAGAIN;
 
 	if (source->depth > MAX_INCLUDE_DEPTH) {
 		error(ap->logopt,
@@ -686,10 +685,8 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 
 			inc = check_self_include(key, ctxt);
 
-			master_source_current_wait(ap->entry);
-			ap->entry->current = source;
-
-			inc_source = prepare_plus_include(ap, age, key, inc, ctxt);
+			inc_source = prepare_plus_include(ap, source,
+							  age, key, inc, ctxt);
 			if (!inc_source) {
 				debug(ap->logopt,
 				      "failed to select included map %s", key);
@@ -728,10 +725,10 @@ int lookup_read_map(struct autofs_point *ap, time_t age, void *context)
 }
 
 static int lookup_one(struct autofs_point *ap,
+		      struct map_source *source,
 		      const char *key, int key_len,
 		      struct lookup_context *ctxt)
 {
-	struct map_source *source;
 	struct mapent_cache *mc;
 	char mkey[KEY_MAX_LEN + 1];
 	char mapent[MAPENT_MAX_LEN + 1];
@@ -740,10 +737,6 @@ static int lookup_one(struct autofs_point *ap,
 	unsigned int k_len, m_len;
 	int entry, ret;
 
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
 	mc = source->mc;
 
 	f = open_fopen_r(ctxt->mapname);
@@ -770,10 +763,8 @@ static int lookup_one(struct autofs_point *ap,
 
 				inc = check_self_include(mkey, ctxt);
 
-				master_source_current_wait(ap->entry);
-				ap->entry->current = source;
-
-				inc_source = prepare_plus_include(ap, age, mkey, inc, ctxt);
+				inc_source = prepare_plus_include(ap, source,
+								  age, mkey, inc, ctxt);
 				if (!inc_source) {
 					debug(ap->logopt,
 					      MODPREFIX
@@ -834,9 +825,9 @@ static int lookup_one(struct autofs_point *ap,
 	return CHE_MISSING;
 }
 
-static int lookup_wild(struct autofs_point *ap, struct lookup_context *ctxt)
+static int lookup_wild(struct autofs_point *ap,
+		       struct map_source *source, struct lookup_context *ctxt)
 {
-	struct map_source *source;
 	struct mapent_cache *mc;
 	char mkey[KEY_MAX_LEN + 1];
 	char mapent[MAPENT_MAX_LEN + 1];
@@ -845,10 +836,6 @@ static int lookup_wild(struct autofs_point *ap, struct lookup_context *ctxt)
 	unsigned int k_len, m_len;
 	int entry, ret;
 
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
 	mc = source->mc;
 
 	f = open_fopen_r(ctxt->mapname);
@@ -892,24 +879,17 @@ static int lookup_wild(struct autofs_point *ap, struct lookup_context *ctxt)
 }
 
 static int check_map_indirect(struct autofs_point *ap,
+			      struct map_source *source,
 			      char *key, int key_len,
 			      struct lookup_context *ctxt)
 {
-	struct map_source *source;
 	struct mapent_cache *mc;
 	struct mapent *exists;
 	int ret = CHE_OK;
 
-	source = ap->entry->current;
-	ap->entry->current = NULL;
-	master_source_current_signal(ap->entry);
-
 	mc = source->mc;
 
-	master_source_current_wait(ap->entry);
-	ap->entry->current = source;
-
-	ret = lookup_one(ap, key, key_len, ctxt);
+	ret = lookup_one(ap, source, key, key_len, ctxt);
 	if (ret == CHE_COMPLETED)
 		return NSS_STATUS_COMPLETED;
 
@@ -932,10 +912,7 @@ static int check_map_indirect(struct autofs_point *ap,
 		struct mapent *we;
 		int wild = CHE_MISSING;
 
-		master_source_current_wait(ap->entry);
-		ap->entry->current = source;
-
-		wild = lookup_wild(ap, ctxt);
+		wild = lookup_wild(ap, source, ctxt);
 		/*
 		 * Check for map change and update as needed for
 		 * following cache lookup.
@@ -1011,8 +988,10 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 				cache_writelock(smc);
 				sme = cache_lookup_distinct(smc, key);
 				/* Negative timeout expired for non-existent entry. */
-				if (sme && !sme->mapent)
-					cache_delete(smc, key);
+				if (sme && !sme->mapent) {
+					if (cache_pop_mapent(sme) == CHE_FAIL)
+						cache_delete(smc, key);
+				}
 				cache_unlock(smc);
 			}
 		}
@@ -1069,10 +1048,8 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		if (!lkp_key)
 			return NSS_STATUS_UNKNOWN;
 
-		master_source_current_wait(ap->entry);
-		ap->entry->current = source;
-
-		status = check_map_indirect(ap, lkp_key, strlen(lkp_key), ctxt);
+		status = check_map_indirect(ap, source,
+					    lkp_key, strlen(lkp_key), ctxt);
 		free(lkp_key);
 		if (status) {
 			if (status == NSS_STATUS_COMPLETED)
@@ -1081,7 +1058,18 @@ int lookup_mount(struct autofs_point *ap, const char *name, int name_len, void *
 		}
 	}
 
-	cache_readlock(mc);
+	/*
+	 * We can't take the writelock for direct mounts. If we're
+	 * starting up or trying to re-connect to an existing direct
+	 * mount we could be iterating through the map entries with
+	 * the readlock held. But we don't need to update the cache
+	 * when we're starting up so just take the readlock in that
+	 * case.
+	 */
+	if (ap->flags & MOUNT_FLAG_REMOUNT)
+		cache_readlock(mc);
+	else
+		cache_writelock(mc);
 do_cache_lookup:
 	me = cache_lookup(mc, key);
 	/*
@@ -1097,11 +1085,21 @@ do_cache_lookup:
 		if (!me)
 			me = cache_lookup_distinct(mc, "*");
 	}
-	if (me && me->mapent && (me->source == source || *me->key == '/')) {
-		pthread_cleanup_push(cache_lock_cleanup, mc);
-		strcpy(mapent_buf, me->mapent);
-		mapent = mapent_buf;
-		pthread_cleanup_pop(0);
+	if (me && me->mapent) {
+		/*
+		 * If this is a lookup add wildcard match for later validation
+		 * checks and negative cache lookups.
+		 */
+		if (!(ap->flags & MOUNT_FLAG_REMOUNT) &&
+		    ap->type == LKP_INDIRECT && *me->key == '*') {
+			ret = cache_update(mc, source, key, me->mapent, me->age);
+			if (!(ret & (CHE_OK | CHE_UPDATED)))
+				me = NULL;
+		}
+		if (me && (me->source == source || *me->key == '/')) {
+			strcpy(mapent_buf, me->mapent);
+			mapent = mapent_buf;
+		}
 	}
 	cache_unlock(mc);
 
@@ -1118,17 +1116,11 @@ do_cache_lookup:
 	ret = ctxt->parse->parse_mount(ap, key, key_len,
 				       mapent, ctxt->parse->context);
 	if (ret) {
-		time_t now = time(NULL);
-		int rv = CHE_OK;
-
+		/* Don't update negative cache when re-connecting */
+		if (ap->flags & MOUNT_FLAG_REMOUNT)
+			return NSS_STATUS_TRYAGAIN;
 		cache_writelock(mc);
-		me = cache_lookup_distinct(mc, key);
-		if (!me)
-			rv = cache_update(mc, source, key, NULL, now);
-		if (rv != CHE_FAIL) {
-			me = cache_lookup_distinct(mc, key);
-			me->status = now + ap->negative_timeout;
-		}
+		cache_update_negative(mc, source, key, ap->negative_timeout);
 		cache_unlock(mc);
 		return NSS_STATUS_TRYAGAIN;
 	}
