@@ -52,6 +52,7 @@
 #include <net/if.h>
 #include <netinet/in.h>
 #include <netdb.h>
+#include <ifaddrs.h>
 
 #include "rpc_subs.h"
 #include "replicated.h"
@@ -72,8 +73,8 @@ static int volatile ifc_last_len = 0;
 /* Get numeric value of the n bits starting at position p */
 #define getbits(x, p, n)	((x >> (p + 1 - n)) & ~(~0 << n))
 
-#define max(x, y)	(x >= y ? x : y)
-#define mmax(x, y, z)	(max(x, y) == x ? max(x, z) : max(y, z))
+#define mymax(x, y)	(x >= y ? x : y)
+#define mmax(x, y, z)	(mymax(x, y) == x ? mymax(x, z) : mymax(y, z))
 
 unsigned int ipv6_mask_cmp(uint32_t *host, uint32_t *iface, uint32_t *mask)
 {
@@ -110,58 +111,18 @@ void seed_random(void)
 	return;
 }
 
-static int alloc_ifreq(struct ifconf *ifc, int sock)
-{
-	int ret, lastlen = ifc_last_len, len = ifc_buf_len;
-	char err_buf[MAX_ERR_BUF], *buf;
-
-	while (1) {
-		buf = malloc(len);
-		if (!buf) {
-			char *estr = strerror_r(errno, err_buf, MAX_ERR_BUF);
-			logerr("malloc: %s", estr);
-			return 0;
-		}
-
-		ifc->ifc_len = len;
-		ifc->ifc_req = (struct ifreq *) buf;
-
-		ret = ioctl(sock, SIOCGIFCONF, ifc);
-		if (ret == -1) {
-			char *estr = strerror_r(errno, err_buf, MAX_ERR_BUF);
-			logerr("ioctl: %s", estr);
-			free(buf);
-			return 0;
-		}
-
-		if (ifc->ifc_len <= lastlen)
-			break;
-
-		lastlen = ifc->ifc_len;
-		len += MAX_IFC_BUF;
-		free(buf);
-	}
-
-	if (lastlen != ifc_last_len) {
-		ifc_last_len = lastlen;
-		ifc_buf_len = len;
-	}
-
-	return 1;
-}
-
 static unsigned int get_proximity(struct sockaddr *host_addr)
 {
+	struct ifaddrs *ifa = NULL;
+	struct ifaddrs *this;
 	struct sockaddr_in *addr, *msk_addr, *if_addr;
 	struct sockaddr_in6 *addr6, *msk6_addr, *if6_addr;
 	struct in_addr *hst_addr;
 	struct in6_addr *hst6_addr;
 	int addr_len;
-	char buf[MAX_ERR_BUF], *ptr;
-	struct ifconf ifc;
-	struct ifreq *ifr, nmptr;
-	int sock, ret, i;
+	char buf[MAX_ERR_BUF];
 	uint32_t mask, ha, ia, *mask6, *ha6, *ia6;
+	int ret;
 
 	addr = NULL;
 	addr6 = NULL;
@@ -170,13 +131,14 @@ static unsigned int get_proximity(struct sockaddr *host_addr)
 	mask6 = NULL;
 	ha6 = NULL;
 	ia6 = NULL;
+	ha = 0;
 
 	switch (host_addr->sa_family) {
 	case AF_INET:
 		addr = (struct sockaddr_in *) host_addr;
 		hst_addr = (struct in_addr *) &addr->sin_addr;
 		ha = ntohl((uint32_t) hst_addr->s_addr);
-		addr_len = sizeof(hst_addr);
+		addr_len = sizeof(*hst_addr);
 		break;
 
 	case AF_INET6:
@@ -186,7 +148,7 @@ static unsigned int get_proximity(struct sockaddr *host_addr)
 		addr6 = (struct sockaddr_in6 *) host_addr;
 		hst6_addr = (struct in6_addr *) &addr6->sin6_addr;
 		ha6 = &hst6_addr->s6_addr32[0];
-		addr_len = sizeof(hst6_addr);
+		addr_len = sizeof(*hst6_addr);
 		break;
 #endif
 
@@ -194,95 +156,74 @@ static unsigned int get_proximity(struct sockaddr *host_addr)
 		return PROXIMITY_ERROR;
 	}
 
-	sock = open_sock(AF_INET, SOCK_DGRAM, 0);
-	if (sock < 0) {
+	ret = getifaddrs(&ifa);
+	if (ret) {
 		char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-		logerr("socket creation failed: %s", estr);
+		logerr("getifaddrs: %s", estr);
 		return PROXIMITY_ERROR;
 	}
 
-	if (!alloc_ifreq(&ifc, sock)) {
-		close(sock);
-		return PROXIMITY_ERROR;
-	}
+	this = ifa;
+	while (this) {
+		if (!(this->ifa_flags & IFF_UP) ||
+		    this->ifa_flags & IFF_POINTOPOINT ||
+		    this->ifa_addr == NULL) {
+			this = this->ifa_next;
+			continue;
+		}
 
-	/* For each interface */
-
-	/* Is the address a local interface */
-	i = 0;
-	ptr = (char *) &ifc.ifc_buf[0];
-
-	while (ptr < (char *) ifc.ifc_req + ifc.ifc_len) {
-		ifr = (struct ifreq *) ptr;
-
-		switch (ifr->ifr_addr.sa_family) {
+		switch (this->ifa_addr->sa_family) {
 		case AF_INET:
 			if (host_addr->sa_family == AF_INET6)
 				break;
-			if_addr = (struct sockaddr_in *) &ifr->ifr_addr;
+			if_addr = (struct sockaddr_in *) this->ifa_addr;
 			ret = memcmp(&if_addr->sin_addr, hst_addr, addr_len);
 			if (!ret) {
-				close(sock);
-				free(ifc.ifc_req);
+				freeifaddrs(ifa);
 				return PROXIMITY_LOCAL;
 			}
 			break;
 
 		case AF_INET6:
-#ifndef WITH_LIBTIRPC
-			return PROXIMITY_UNSUPPORTED;
-#else
+#ifdef WITH_LIBTIRPC
 			if (host_addr->sa_family == AF_INET)
 				break;
-
-			if6_addr = (struct sockaddr_in6 *) &ifr->ifr_addr;
+			if6_addr = (struct sockaddr_in6 *) this->ifa_addr;
 			ret = memcmp(&if6_addr->sin6_addr, hst6_addr, addr_len);
 			if (!ret) {
-				close(sock);
-				free(ifc.ifc_req);
+				freeifaddrs(ifa);
 				return PROXIMITY_LOCAL;
 			}
 #endif
-
 		default:
 			break;
 		}
-
-		i++;
-		ptr = (char *) &ifc.ifc_req[i];
+		this = this->ifa_next;
 	}
 
-	i = 0;
-	ptr = (char *) &ifc.ifc_buf[0];
-
-	while (ptr < (char *) ifc.ifc_req + ifc.ifc_len) {
-		ifr = (struct ifreq *) ptr;
-
-		nmptr = *ifr;
-		ret = ioctl(sock, SIOCGIFNETMASK, &nmptr);
-		if (ret == -1) {
-			char *estr = strerror_r(errno, buf, MAX_ERR_BUF);
-			logerr("ioctl: %s", estr);
-			close(sock);
-			free(ifc.ifc_req);
-			return PROXIMITY_ERROR;
+	this = ifa;
+	while (this) {
+		if (!(this->ifa_flags & IFF_UP) ||
+		    this->ifa_flags & IFF_POINTOPOINT ||
+		    this->ifa_addr == NULL) {
+			this = this->ifa_next;
+			continue;
 		}
 
-		switch (ifr->ifr_addr.sa_family) {
+		switch (this->ifa_addr->sa_family) {
 		case AF_INET:
 			if (host_addr->sa_family == AF_INET6)
 				break;
-			if_addr = (struct sockaddr_in *) &ifr->ifr_addr;
+			if_addr = (struct sockaddr_in *) this->ifa_addr;
 			ia =  ntohl((uint32_t) if_addr->sin_addr.s_addr);
 
-			/* Is the address within a localiy attached subnet */
+			/* Is the address within a localy attached subnet */
 
-			msk_addr = (struct sockaddr_in *) &nmptr.ifr_netmask;
+			msk_addr = (struct sockaddr_in *) this->ifa_netmask;
 			mask = ntohl((uint32_t) msk_addr->sin_addr.s_addr);
 
 			if ((ia & mask) == (ha & mask)) {
-				close(sock);
-				free(ifc.ifc_req);
+				freeifaddrs(ifa);
 				return PROXIMITY_SUBNET;
 			}
 
@@ -304,55 +245,45 @@ static unsigned int get_proximity(struct sockaddr *host_addr)
 				break;
 
 			if ((ia & mask) == (ha & mask)) {
-				close(sock);
-				free(ifc.ifc_req);
+				freeifaddrs(ifa);
 				return PROXIMITY_NET;
 			}
 			break;
 
 		case AF_INET6:
-#ifndef WITH_LIBTIRPC
-			return PROXIMITY_UNSUPPORTED;
-#else
+#ifdef WITH_LIBTIRPC
 			if (host_addr->sa_family == AF_INET)
 				break;
-
-			if6_addr = (struct sockaddr_in6 *) &ifr->ifr_addr;
+			if6_addr = (struct sockaddr_in6 *) this->ifa_addr;
 			ia6 = &if6_addr->sin6_addr.s6_addr32[0];
 
 			/* Is the address within the network of the interface */
 
-			msk6_addr = (struct sockaddr_in6 *) &nmptr.ifr_netmask;
+			msk6_addr = (struct sockaddr_in6 *) this->ifa_netmask;
 			mask6 = &msk6_addr->sin6_addr.s6_addr32[0];
 
 			if (ipv6_mask_cmp(ha6, ia6, mask6)) {
-				close(sock);
-				free(ifc.ifc_req);
+				freeifaddrs(ifa);
 				return PROXIMITY_SUBNET;
 			}
 
 			/* How do we define "local network" in ipv6? */
 #endif
-			break;
-
 		default:
 			break;
 		}
-
-		i++;
-		ptr = (char *) &ifc.ifc_req[i];
+		this = this->ifa_next;
 	}
 
-	close(sock);
-	free(ifc.ifc_req);
+	freeifaddrs(ifa);
 
 	return PROXIMITY_OTHER;
 }
 
-static struct host *new_host(const char *name,
-			     struct sockaddr *addr, size_t addr_len,
-			     unsigned int proximity, unsigned int weight,
-			     unsigned int options)
+struct host *new_host(const char *name,
+		      struct sockaddr *addr, size_t addr_len,
+		      unsigned int proximity, unsigned int weight,
+		      unsigned int options)
 {
 	struct host *new;
 	struct sockaddr *tmp2;
@@ -490,7 +421,7 @@ void free_host_list(struct host **list)
 
 static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 			 struct conn_info *pm_info, struct conn_info *rpc_info,
-			 const char *proto, unsigned int version, int port)
+			 int proto, unsigned int version, int port)
 {
 	unsigned int random_selection = host->options & MOUNT_FLAG_RANDOM_SELECT;
 	unsigned int use_weight_only = host->options & MOUNT_FLAG_USE_WEIGHT_ONLY;
@@ -504,31 +435,31 @@ static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 	int status, count = 0;
 
 	if (host->addr)
-		debug(logopt, "called with host %s(%s) proto %s version 0x%x",
+		debug(logopt, "called with host %s(%s) proto %d version 0x%x",
 		      host->name, get_addr_string(host->addr, buf, len),
 		      proto, version);
 	else
 		debug(logopt,
-		      "called for host %s proto %s version 0x%x",
+		      "called for host %s proto %d version 0x%x",
 		      host->name, proto, version);
 
-	rpc_info->proto = getprotobyname(proto);
-	if (!rpc_info->proto)
-		return 0;
+	rpc_info->proto = proto;
+	if (port < 0) {
+		if (version & NFS4_REQUESTED)
+			rpc_info->port = NFS_PORT;
+		else
+			port = 0;
+	} else if (port > 0)
+		rpc_info->port = port;
 
 	memset(&parms, 0, sizeof(struct pmap));
-
 	parms.pm_prog = NFS_PROGRAM;
-	parms.pm_prot = rpc_info->proto->p_proto;
+	parms.pm_prot = proto;
 
 	if (!(version & NFS4_REQUESTED))
 		goto v3_ver;
 
-	if (port < 0)
-		rpc_info->port = NFS_PORT;
-	else if (port > 0)
-		rpc_info->port = port;
-	else {
+	if (!port) {
 		status = rpc_portmap_getclient(pm_info,
 				host->name, host->addr, host->addr_len,
 				proto, RPC_CLOSE_DEFAULT);
@@ -550,7 +481,7 @@ static unsigned int get_nfs_info(unsigned logopt, struct host *host,
 		}
 	}
 
-	if (rpc_info->proto->p_proto == IPPROTO_UDP)
+	if (rpc_info->proto == IPPROTO_UDP)
 		status = rpc_udp_getclient(rpc_info, NFS_PROGRAM, NFS4_VERSION);
 	else
 		status = rpc_tcp_getclient(rpc_info, NFS_PROGRAM, NFS4_VERSION);
@@ -588,7 +519,7 @@ v3_ver:
 	if (!(version & NFS3_REQUESTED))
 		goto v2_ver;
 
-	if (port <= 0 && !(version & NFS4_REQUESTED && port == 0)) {
+	if (!port && !pm_info->client) {
 		status = rpc_portmap_getclient(pm_info,
 				host->name, host->addr, host->addr_len,
 				proto, RPC_CLOSE_DEFAULT);
@@ -599,9 +530,7 @@ v3_ver:
 			goto done_ver;
 	}
 
-	if (port > 0)
-		rpc_info->port = port;
-	else {
+	if (!port) {
 		parms.pm_vers = NFS3_VERSION;
 		status = rpc_portmap_getport(pm_info, &parms, &rpc_info->port);
 		if (status == -EHOSTUNREACH || status == -ETIMEDOUT) {
@@ -611,7 +540,7 @@ v3_ver:
 			goto v2_ver;
 	}
 
-	if (rpc_info->proto->p_proto == IPPROTO_UDP)
+	if (rpc_info->proto == IPPROTO_UDP)
 		status = rpc_udp_getclient(rpc_info, NFS_PROGRAM, NFS3_VERSION);
 	else
 		status = rpc_tcp_getclient(rpc_info, NFS_PROGRAM, NFS3_VERSION);
@@ -646,9 +575,18 @@ v2_ver:
 	if (!(version & NFS2_REQUESTED))
 		goto done_ver;
 
-	if (port > 0)
-		rpc_info->port = port;
-	else {
+	if (!port && !pm_info->client) {
+		status = rpc_portmap_getclient(pm_info,
+				host->name, host->addr, host->addr_len,
+				proto, RPC_CLOSE_DEFAULT);
+		if (status == -EHOSTUNREACH) {
+			supported = status;
+			goto done_ver;
+		} else if (status)
+			goto done_ver;
+	}
+
+	if (!port) {
 		parms.pm_vers = NFS2_VERSION;
 		status = rpc_portmap_getport(pm_info, &parms, &rpc_info->port);
 		if (status == -EHOSTUNREACH || status == -ETIMEDOUT) {
@@ -658,7 +596,7 @@ v2_ver:
 			goto done_ver;
 	}
 
-	if (rpc_info->proto->p_proto == IPPROTO_UDP)
+	if (rpc_info->proto == IPPROTO_UDP)
 		status = rpc_udp_getclient(rpc_info, NFS_PROGRAM, NFS2_VERSION);
 	else
 		status = rpc_tcp_getclient(rpc_info, NFS_PROGRAM, NFS2_VERSION);
@@ -689,7 +627,7 @@ v2_ver:
 	}
 
 done_ver:
-	if (rpc_info->proto->p_proto == IPPROTO_UDP) {
+	if (rpc_info->proto == IPPROTO_UDP) {
 		rpc_destroy_udp_client(rpc_info);
 		rpc_destroy_udp_client(pm_info);
 	} else {
@@ -746,7 +684,7 @@ static int get_vers_and_cost(unsigned logopt, struct host *host,
 
 	if (version & TCP_REQUESTED) {
 		supported = get_nfs_info(logopt, host,
-				   &pm_info, &rpc_info, "tcp", vers, port);
+				   &pm_info, &rpc_info, IPPROTO_TCP, vers, port);
 		if (IS_ERR(supported)) {
 			if (ERR(supported) == EHOSTUNREACH ||
 			    ERR(supported) == ETIMEDOUT)
@@ -759,7 +697,7 @@ static int get_vers_and_cost(unsigned logopt, struct host *host,
 
 	if (version & UDP_REQUESTED) {
 		supported = get_nfs_info(logopt, host,
-				   &pm_info, &rpc_info, "udp", vers, port);
+				   &pm_info, &rpc_info, IPPROTO_UDP, vers, port);
 		if (IS_ERR(supported)) {
 			if (!ret && ERR(supported) == ETIMEDOUT)
 				return ret;
@@ -780,7 +718,7 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 	socklen_t len = INET6_ADDRSTRLEN;
 	char buf[len + 1];
 	struct conn_info pm_info, rpc_info;
-	const char *proto;
+	int proto;
 	unsigned int vers;
 	struct timeval start, end;
 	struct timezone tz;
@@ -819,10 +757,10 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 	 *  So, we do the conversion here.
 	 */
 	if (version & UDP_SELECTED_MASK) {
-		proto = "udp";
+		proto = IPPROTO_UDP;
 		version >>= 8;
 	} else
-		proto = "tcp";
+		proto = IPPROTO_TCP;
 
 	switch (version) {
 	case NFS2_SUPPORTED:
@@ -839,9 +777,7 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 		return 0;
 	}
 
-	rpc_info.proto = getprotobyname(proto);
-	if (!rpc_info.proto)
-		return 0;
+	rpc_info.proto = proto;
 
 	if (port > 0)
 		rpc_info.port = port;
@@ -857,14 +793,14 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 
 		memset(&parms, 0, sizeof(struct pmap));
 		parms.pm_prog = NFS_PROGRAM;
-		parms.pm_prot = rpc_info.proto->p_proto;
+		parms.pm_prot = rpc_info.proto;
 		parms.pm_vers = vers;
 		ret = rpc_portmap_getport(&pm_info, &parms, &rpc_info.port);
 		if (ret < 0)
 			goto done;
 	}
 
-	if (rpc_info.proto->p_proto == IPPROTO_UDP)
+	if (rpc_info.proto == IPPROTO_UDP)
 		status = rpc_udp_getclient(&rpc_info, NFS_PROGRAM, vers);
 	else
 		status = rpc_tcp_getclient(&rpc_info, NFS_PROGRAM, vers);
@@ -886,7 +822,7 @@ static int get_supported_ver_and_cost(unsigned logopt, struct host *host,
 		}
 	}
 done:
-	if (rpc_info.proto->p_proto == IPPROTO_UDP) {
+	if (rpc_info.proto == IPPROTO_UDP) {
 		rpc_destroy_udp_client(&rpc_info);
 		rpc_destroy_udp_client(&pm_info);
 	} else {
@@ -1025,7 +961,7 @@ int prune_host_list(unsigned logopt, struct host **list,
 
 	max_tcp_count = mmax(v4_tcp_count, v3_tcp_count, v2_tcp_count);
 	max_udp_count = mmax(v4_udp_count, v3_udp_count, v2_udp_count);
-	max_count = max(max_tcp_count, max_udp_count);
+	max_count = mymax(max_tcp_count, max_udp_count);
 
 	if (max_count == v4_tcp_count) {
 		selected_version = NFS4_TCP_SUPPORTED;
@@ -1105,28 +1041,19 @@ static int add_new_host(struct host **list,
 	unsigned int prx;
 	int addr_len;
 
+	prx = get_proximity(host_addr->ai_addr);
+
 	/*
-	 * If we are using random selection we pretend all hosts are at
-	 * the same proximity so hosts further away don't get excluded.
-	 * We can't use PROXIMITY_LOCAL or we won't perform an RPC ping
-	 * to remove hosts that may be down.
+	 * If we want the weight to be the determining factor
+	 * when selecting a host, or we are using random selection,
+	 * then all hosts must have the same proximity. However,
+	 * if this is the local machine it should always be used
+	 * since it is certainly available.
 	 */
-	if (!host_addr)
+	if (prx != PROXIMITY_LOCAL &&
+	   (options & (MOUNT_FLAG_USE_WEIGHT_ONLY |
+		       MOUNT_FLAG_RANDOM_SELECT)))
 		prx = PROXIMITY_SUBNET;
-	else {
-		prx = get_proximity(host_addr->ai_addr);
-		/*
-		 * If we want the weight to be the determining factor
-		 * when selecting a host, or we are using random selection,
-		 * then all hosts must have the same proximity. However,
-		 * if this is the local machine it should always be used
-		 * since it is certainly available.
-		 */
-		if (prx != PROXIMITY_LOCAL &&
-		   (options & (MOUNT_FLAG_USE_WEIGHT_ONLY |
-			       MOUNT_FLAG_RANDOM_SELECT)))
-			prx = PROXIMITY_SUBNET;
-	}
 
 	/*
 	 * If we tried to add an IPv6 address and we don't have IPv6
@@ -1223,7 +1150,7 @@ try_name:
 				rr4++;
 		} else if (this->ai_family == AF_INET6) {
 			struct sockaddr_in6 *addr = (struct sockaddr_in6 *) this->ai_addr;
-			if (!IN6_IS_ADDR_LOOPBACK(addr->sin6_addr.__in6_u.__u6_addr32))
+			if (!IN6_IS_ADDR_LOOPBACK(addr->sin6_addr.s6_addr32))
 				rr6++;
 		}
 		this = this->ai_next;
